import enum

from tutubo.pytube.contrib.search import Search as _Search
from tutubo.models import *
from tutubo.ytmus import *


class SearchType(enum.IntEnum):
    YOUTUBE = enum.auto()
    VIDEOS = enum.auto()
    RELATED_VIDEOS = enum.auto()
    CHANNELS = enum.auto()
    PLAYLISTS = enum.auto()
    YOUTUBE_MIX = enum.auto()
    RELATED_QUERIES = enum.auto()

    MUSIC = enum.auto()
    MUSIC_TRACK = enum.auto()
    MUSIC_VIDEO = enum.auto()
    MUSIC_ARTIST = enum.auto()
    MUSIC_PLAYLIST = enum.auto()
    MUSIC_ALBUM = enum.auto()

    ALL = enum.auto()


class YoutubeSearch(_Search):
    def __init__(self, query, preview=True, thumbnail_url=""):
        super().__init__(query)
        self.preview = preview
        self.thumbnail_url = thumbnail_url

    @property
    def as_dict(self):
        return {'query': self.query,
                'image': self.thumbnail_url}

    # pytube extensions - TODO https://github.com/pytube/pytube/pull/1133
    def iterate_youtube(self, max_res=-1, search_type=SearchType.YOUTUBE):
        idx = 0
        for r in self._iterate_and_parse(search_type=search_type):
            idx += 1
            if 0 < max_res < idx:
                break
            if not self.preview:
                r = r.get()
            yield r

    def _iterate_and_parse(self, continuation=None,
                           search_type=SearchType.ALL):
        """Fetch from the innertube API and parse the results.

        :param str continuation:
            Continuation string for fetching results.
        :rtype: tuple
        :returns:
            A tuple of a list of YouTube objects and a continuation string.
        """
        # Begin by executing the query and identifying the relevant sections
        #  of the results
        raw_results = self.fetch_query(continuation)

        # Initial result is handled by try block, continuations by except block
        try:
            sections = \
                raw_results['contents']['twoColumnSearchResultsRenderer'][
                    'primaryContents']['sectionListRenderer']['contents']
        except KeyError:
            sections = raw_results['onResponseReceivedCommands'][0][
                'appendContinuationItemsAction']['continuationItems']
        item_renderer = None
        continuation_renderer = None
        for s in sections:
            if 'itemSectionRenderer' in s:
                item_renderer = s['itemSectionRenderer']
            if 'continuationItemRenderer' in s:
                continuation_renderer = s['continuationItemRenderer']

        # If the continuationItemRenderer doesn't exist, assume no further results
        if continuation_renderer:
            next_continuation = continuation_renderer['continuationEndpoint'][
                'continuationCommand']['token']
        else:
            next_continuation = None

        # If the itemSectionRenderer doesn't exist, assume no results.
        if item_renderer:
            raw_video_list = item_renderer['contents']
            for video_details in raw_video_list:
                # Skip over ads
                if video_details.get('searchPyvRenderer', {}).get('ads', None):
                    continue

                # Skip "recommended" type videos e.g. "people also watched" and "popular X"
                #  that break up the search results
                elif 'shelfRenderer' in video_details and \
                        search_type in [SearchType.ALL,
                                        SearchType.YOUTUBE,
                                        SearchType.RELATED_VIDEOS]:
                    content = video_details['shelfRenderer']['content'].get("verticalListRenderer")
                    if content:
                        for v in content['items']:
                            yield RelatedVideoPreview(v['videoRenderer'])
                        continue

                # Skip auto-generated "mix" playlist results
                elif 'radioRenderer' in video_details and \
                        search_type in [SearchType.ALL,
                                        SearchType.YOUTUBE,
                                        SearchType.YOUTUBE_MIX]:
                    yield YoutubeMixPreview(video_details['radioRenderer'])
                    continue

                # Skip playlist results
                elif 'playlistRenderer' in video_details and \
                        search_type in [SearchType.ALL,
                                        SearchType.YOUTUBE,
                                        SearchType.PLAYLISTS]:
                    yield PlaylistPreview(video_details['playlistRenderer'])
                    continue

                # Skip channel results
                elif 'channelRenderer' in video_details and \
                        search_type in [SearchType.ALL,
                                        SearchType.YOUTUBE,
                                        SearchType.CHANNELS]:
                    yield ChannelPreview(video_details['channelRenderer'])
                    continue

                # Skip 'people also searched for' results
                elif 'horizontalCardListRenderer' in video_details and \
                        search_type in [SearchType.RELATED_QUERIES,
                                        SearchType.YOUTUBE,
                                        SearchType.ALL]:
                    for v in video_details['horizontalCardListRenderer'][
                        'cards']:
                        yield RelatedSearch(v['searchRefinementCardRenderer'])
                    continue

                # Can't seem to reproduce, probably related to typo fix suggestions
                elif 'didYouMeanRenderer' in video_details:
                    continue

                # Seems to be the renderer used for the image shown on a no results page
                elif 'backgroundPromoRenderer' in video_details:
                    continue

                # no more results
                elif 'messageRenderer' in video_details:
                    return

                elif 'videoRenderer' in video_details and \
                        search_type in [SearchType.ALL,
                                        SearchType.YOUTUBE,
                                        SearchType.VIDEOS]:
                    yield VideoPreview(video_details['videoRenderer'])

        if next_continuation:
            for r in self._iterate_and_parse(next_continuation, search_type):
                yield r

    def iterate_videos(self, max_res=-1):
        for v in self.iterate_youtube(max_res):
            if isinstance(v, Video) or isinstance(v, VideoPreview):
                yield v

    def iterate_related_videos(self, max_res=-1):
        for v in self.iterate_youtube(max_res):
            if isinstance(v, RelatedVideo):
                yield v

    def iterate_channels(self, max_res=-1):
        for v in self.iterate_youtube(max_res):
            if isinstance(v, Channel) or isinstance(v, ChannelPreview):
                yield v

    def iterate_playlists(self, max_res=-1):
        for v in self.iterate_youtube(max_res):
            if isinstance(v, PlaylistPreview) or isinstance(v, Playlist):
                yield v

    def iterate_mixes(self, max_res=-1):
        for v in self.iterate_youtube(max_res):
            if isinstance(v, YoutubeMixPreview):
                yield v

    def iterate_queries(self, max_res=-1):
        for v in self.iterate_youtube(max_res):
            if isinstance(v, RelatedSearch):
                yield v

    # youtube music
    def iterate_youtube_music(self, search_type=SearchType.MUSIC):
        ytmusic = YTMusic()
        for r in ytmusic.search(self.query):
            if r["resultType"] == "video" and \
                    search_type in [SearchType.MUSIC,
                                    SearchType.ALL,
                                    SearchType.MUSIC_VIDEO]:
                yield MusicVideo(r)
            elif r["resultType"] == "song" and \
                    search_type in [SearchType.MUSIC,
                                    SearchType.ALL,
                                    SearchType.MUSIC_TRACK]:
                yield MusicTrack(r)
            elif r["resultType"] == "album" and \
                    search_type in [SearchType.MUSIC,
                                    SearchType.ALL,
                                    SearchType.MUSIC_ALBUM]:
                try:
                    a = ytmusic.get_album(r["browseId"])
                except:
                    continue
                r.update(a)
                yield MusicAlbum(r)
            elif r["resultType"] == "playlist" and \
                    search_type in [SearchType.MUSIC,
                                    SearchType.ALL,
                                    SearchType.MUSIC_PLAYLIST]:
                try:
                    a = ytmusic.get_playlist(r["browseId"])
                except:
                    continue
                r.update(a)
                yield MusicPlaylist(r)
            elif r["resultType"] == "artist" and \
                    search_type in [SearchType.MUSIC,
                                    SearchType.MUSIC_ARTIST,
                                    SearchType.ALL]:
                try:
                    a = ytmusic.get_artist(r["browseId"])
                except:
                    continue
                r.update(a)
                yield MusicArtist(r)

    def iterate_music_albums(self, max_res=-1):
        n = 0
        for v in self.iterate_youtube_music(SearchType.MUSIC_ALBUM):
            yield v
            n += 1
            if n > max_res:
                break

    def iterate_music_artists(self, max_res=-1):
        n = 0
        for v in self.iterate_youtube_music(SearchType.MUSIC_ARTIST):
            yield v
            n += 1
            if n > max_res:
                break

    def iterate_music_tracks(self, max_res=-1):
        n = 0
        for v in self.iterate_youtube_music(SearchType.MUSIC_TRACK):
            yield v
            n += 1
            if n > max_res:
                break

    def iterate_music_playlists(self, max_res=-1):
        n = 0
        for v in self.iterate_youtube_music(SearchType.MUSIC_PLAYLIST):
            yield v
            n += 1
            if n > max_res:
                break

    def iterate_music_videos(self, max_res=-1):
        n = 0
        for v in self.iterate_youtube_music(SearchType.MUSIC_VIDEO):
            yield v
            n += 1
            if n > max_res:
                break


def search_yt(query, as_dict=True, parse=False, max_res=50):
    s = YoutubeSearch(query, preview=not parse)
    for v in s.iterate_youtube(max_res=max_res):
        if as_dict:
            yield v.as_dict
        else:
            yield v



from setuptools import setup
import os

PKG_NAME = 'tutubo'

def get_version():
    """ Find the version of this skill"""
    version_file = os.path.join(os.path.dirname(__file__), PKG_NAME, 'version.py')
    major, minor, build, alpha = (None, None, None, None)
    with open(version_file) as f:
        for line in f:
            if 'VERSION_MAJOR' in line:
                major = line.split('=')[1].strip()
            elif 'VERSION_MINOR' in line:
                minor = line.split('=')[1].strip()
            elif 'VERSION_BUILD' in line:
                build = line.split('=')[1].strip()
            elif 'VERSION_ALPHA' in line:
                alpha = line.split('=')[1].strip()

            if ((major and minor and build and alpha) or
                    '# END_VERSION_BLOCK' in line):
                break
    version = f"{major}.{minor}.{build}"
    if int(alpha):
        version += f"a{alpha}"
    return version


setup(
    name=PKG_NAME,
    version=get_version(),
    packages=[PKG_NAME],
    url='https://github.com/OpenJarbas/tutubo',
    license='Apache',
    author='jarbasAI',
    install_requires=["bs4", "requests", "pytube", "ytmusicapi"],
    author_email='jarbasai@mailfence.com',
    description='wrapper around pytube with some new functionality'
)



import json
import time
from ytmusicapi import YTMusic

from tutubo.models import YoutubePreview, Video

_YTMUS = None


def _get_ytmus(max_retries=5):
    """ attempt at avoiding rate limiting errors, avoid creating YTMusic objects unnecessarily

      for v in search_yt_music(phrase, as_dict=False):
      File "/usr/lib/python3.10/site-packages/tutubo/ytmus.py", line 127, in search_yt_music
        ytmusic = YTMusic()
      File "/usr/lib/python3.10/site-packages/ytmusicapi/ytmusic.py", line 96, in __init__
        self.headers.update(get_visitor_id(self._send_get_request))
      File "/usr/lib/python3.10/site-packages/ytmusicapi/helpers.py", line 63, in get_visitor_id
        response = request_func(YTM_DOMAIN)
      File "/usr/lib/python3.10/site-packages/ytmusicapi/ytmusic.py", line 146, in _send_get_request
        response = requests.get(url, params, headers=self.headers, proxies=self.proxies)
      File "/usr/lib/python3.10/site-packages/requests/api.py", line 75, in get
        return request('get', url, params=params, **kwargs)
      (...)
      File "/usr/lib/python3.10/site-packages/requests/adapters.py", line 519, in send
        raise ConnectionError(e, request=request)
    requests.exceptions.ConnectionError: HTTPSConnectionPool(host='music.youtube.com', port=443): Max retries exceeded with url: / (Caused by NewConnectionError('<urllib3.connection.HTTPSConnection object at 0x7f940cece0>: Failed to establish a new connection: [Errno -2] Name or service not known'))

    """
    global _YTMUS
    if _YTMUS:
        return _YTMUS
    for i in range(max_retries):
        try:
            _YTMUS = YTMusic()
            break
        except:  # rate limited
            time.sleep(0.5 * (i + 1))
            continue
    return _YTMUS


class YTMusicResult(YoutubePreview):
    @property
    def title(self):
        return self._raw_data.get("title")

    @property
    def thumbnail_url(self):
        img = self._raw_data.get("image")
        if not img and self._raw_data.get("thumbnails"):
            img = self._raw_data["thumbnails"][-1]["url"]
        return img

    @property
    def artist(self):
        artist = self._raw_data.get("artist")
        if not artist and self._raw_data.get("artists"):
            artist = ", ".join(a["name"] for a in self._raw_data['artists'])
        return artist

    @property
    def description(self):
        return self._raw_data.get("description")

    @property
    def as_dict(self):
        return self._raw_data

    def __dict__(self):
        return self.as_dict

    def __str__(self):
        return json.dumps(self.as_dict, sort_keys=True)


class MusicTrack(YTMusicResult):
    @property
    def watch_url(self):
        return "https://music.youtube.com/watch?v=" + self._raw_data["videoId"]

    @property
    def length(self):
        # converted to seconds or None
        dur = self._raw_data.get("duration")
        if isinstance(dur, str):
            dur = dur.split(":")
            if len(dur) == 2:
                m, s = dur
                return 60 * int(m) + int(s)
            elif len(dur) == 3:
                h, m, s = dur
                return 60 * 60 * int(h) + 60 * int(m) + int(s)
        return None

    @property
    def album(self):
        return self._raw_data.get("album", {}).get("name")

    @property
    def as_dict(self):
        return {"title": self.title,
                "artist": self.artist,
                "image": self.thumbnail_url,
                "url": self.watch_url,
                "duration": self.length}


class MusicVideo(MusicTrack):
    @property
    def watch_url(self):
        return "https://www.youtube.com/watch?v=" + self._raw_data["videoId"]

    def get(self):
        return Video(self.watch_url)


class MusicPlaylist(YTMusicResult):
    @property
    def tracks(self):
        if "tracks" in self._raw_data:
            return [
                MusicTrack(t) for t in self._raw_data["tracks"]
                if t.get("videoId")
            ]
        elif "songs" in self._raw_data:
            return [
                MusicTrack(t) for t in self._raw_data["songs"].get("results", [])
                if t.get("videoId")
            ]
        return []

    @property
    def as_dict(self):
        return {
            "title": self.title,
            "artist": self.artist,
            "image": self.thumbnail_url,
            "playlist": [t.as_dict for t in self.tracks]
        }


class MusicAlbum(MusicPlaylist):
    @property
    def name(self):
        return self.title


class MusicArtist(MusicPlaylist):
    @property
    def name(self):
        return self.artist

    @property
    def as_dict(self):
        return {
            "artist": self.name,
            "image": self.thumbnail_url,
            "playlist": [t.as_dict for t in self.tracks]
        }


def search_yt_music(query, as_dict=True, n_retries=3):
    ytmusic = _get_ytmus(n_retries)
    for r in ytmusic.search(query):
        if r["resultType"] == "video":
            if as_dict:
                yield MusicVideo(r).as_dict
            else:
                yield MusicVideo(r)
        elif r["resultType"] == "song":
            yield MusicTrack(r)
        elif r["resultType"] == "album":
            try:
                a = ytmusic.get_album(r["browseId"])
            except:
                continue
            r.update(a)
            if as_dict:
                yield MusicAlbum(r).as_dict
            else:
                yield MusicAlbum(r)
        elif r["resultType"] == "playlist":
            try:
                a = ytmusic.get_playlist(r["browseId"])
            except:
                continue
            r.update(a)
            if as_dict:
                yield MusicPlaylist(r).as_dict
            else:
                yield MusicPlaylist(r)
        elif r["resultType"] == "artist":
            try:
                a = ytmusic.get_artist(r["browseId"])
            except:
                continue
            r.update(a)
            if as_dict:
                yield MusicArtist(r).as_dict
            else:
                yield MusicArtist(r)



from tutubo.pytube import YouTube as Video, Channel, Playlist


class YoutubePreview:
    def __init__(self, renderer_data):
        self._raw_data = renderer_data


class PlaylistPreview(YoutubePreview):
    def get(self):
        return Playlist(self.playlist_url)

    @property
    def title(self):
        return self._raw_data["title"]['simpleText']

    @property
    def playlist_id(self):
        return self._raw_data["playlistId"]

    @property
    def playlist_url(self):
        return f"https://www.youtube.com/playlist?list={self.playlist_id}"

    @property
    def video_count(self):
        return self._raw_data['videoCount']

    @property
    def featured_videos(self):
        videos = []
        for v in self._raw_data['videos']:
            v = v['childVideoRenderer']
            videos.append({
                "videoId": v['videoId'],
                "url": f"https://youtube.com/watch?v={v['videoId']}",
                "image": f"https://img.youtube.com/vi/{v['videoId']}/default.jpg",
                "title": v["title"]["simpleText"]
            })
        return videos

    @property
    def thumbnail_url(self):
        return self.thumbnails[-1]["url"]

    @property
    def thumbnails(self):
        return [t['thumbnails'][0] for t in self._raw_data['thumbnails']]

    def __str__(self):
        return self.title

    @property
    def as_dict(self):
        return {'playlistId': self.playlist_id,
                'title': self.title,
                'url': self.playlist_url,
                "image": self.thumbnail_url,
                'featured_videos': self.featured_videos}


class YoutubeMixPreview(PlaylistPreview):
    @property
    def thumbnail_url(self):
        return self.thumbnails[-1]["url"]

    @property
    def thumbnails(self):
        return self._raw_data['thumbnail']['thumbnails']

    @property
    def as_dict(self):
        return {'playlistId': self.playlist_id,
                'title': self.title,
                'url': self.playlist_url,
                "image": self.thumbnail_url,
                'featured_videos': self.featured_videos}


class ChannelPreview(YoutubePreview):

    def get(self):
        return Channel(self.channel_url)

    @property
    def title(self):
        return self._raw_data["title"]['simpleText']

    @property
    def description(self):
        return "".join(r["text"] for r in
                       self._raw_data['descriptionSnippet']['runs'])

    @property
    def channel_id(self):
        return self._raw_data["channelId"]

    @property
    def channel_url(self):
        return f"https://www.youtube.com/channel/{self.channel_id}"

    @property
    def video_count(self):
        return int(self._raw_data['videoCountText']["runs"][0]["text"])

    @property
    def thumbnail_url(self):
        return self.thumbnails[-1]["url"]

    @property
    def thumbnails(self):
        return self._raw_data['thumbnail']['thumbnails']

    def __str__(self):
        return self.title

    @property
    def as_dict(self):
        return {'channelId': self.channel_id,
                'title': self.title,
                'image': self.thumbnail_url,
                'url': self.channel_url}


class VideoPreview(YoutubePreview):

    def get(self):
        return Video(self.watch_url)

    @property
    def title(self):
        return "".join(r["text"] for r in
                       self._raw_data['title']['runs'])

    @property
    def author(self):
        return "".join(r["text"] for r in
                       self._raw_data['ownerText']['runs'])

    @property
    def channel_id(self):
        return self._raw_data['ownerText']['runs'][0][
            'navigationEndpoint']['commandMetadata'][
            'webCommandMetadata']['url']

    @property
    def channel_url(self):
        return f'https://www.youtube.com/channel/{self.channel_id}'

    @property
    def video_id(self):
        return self._raw_data['videoId']

    @property
    def watch_url(self):
        return f'https://www.youtube.com/watch?v={self.video_id}'

    @property
    def view_count(self):
        # Livestreams have "runs", non-livestreams have "simpleText",
        #  and scheduled releases do not have 'viewCountText'
        if 'viewCountText' in self._raw_data:
            if 'runs' in self._raw_data['viewCountText']:
                vid_view_count_text = \
                    self._raw_data['viewCountText']['runs'][0][
                        'text']
            else:
                vid_view_count_text = \
                    self._raw_data['viewCountText']['simpleText']
            # Strip ' views' text, then remove commas
            stripped_text = vid_view_count_text.split()[0].replace(
                ',', '')
            if stripped_text != 'No':
                return int(stripped_text)
        return 0

    @property
    def length(self):
        if 'lengthText' in self._raw_data:
            pts = self._raw_data['lengthText']['simpleText'].split(":")
            h, m, s = 0, 0, 0
            if len(pts) == 3:
                h, m, s = pts
            elif len(pts) == 2:
                m, s = pts
            elif len(pts) == 1:
                s = pts
            return int(s) + \
                60 * int(m) + \
                60 * 60 * int(h)
        return 0

    @property
    def is_live(self) -> bool:
        return self.length == 0

    @property
    def thumbnail_url(self):
        return f"https://img.youtube.com/vi/{self.video_id}/default.jpg"

    @property
    def keywords(self):
        return []  # just for api compatibility, requires parsing url!

    def __str__(self):
        return self.title

    @property
    def as_dict(self):
        return {'length': self.length,
                'keywords': self.keywords,
                'image': self.thumbnail_url,
                'title': self.title,
                "author": self.author,
                'url': self.watch_url,
                'videoId': self.video_id}


class RelatedVideoPreview(VideoPreview):
    @property
    def as_dict(self):
        return {'length': self.length,
                'keywords': self.keywords,
                'image': self.thumbnail_url,
                'title': self.title,
                "author": self.author,
                'url': self.watch_url,
                'videoId': self.video_id}


class RelatedSearch(YoutubePreview):

    def get(self, preview=True):
        from tutubo.search import YoutubeSearch
        return YoutubeSearch(self.query, preview=preview)

    @property
    def query(self):
        return "".join(r["text"] for r in self._raw_data['query']['runs'])

    @property
    def thumbnail_url(self):
        return self.thumbnails[-1]["url"]

    @property
    def thumbnails(self):
        return self._raw_data['thumbnail']['thumbnails']

    def __str__(self):
        return self.query

    @property
    def as_dict(self):
        return {'query': self.query,
                'image': self.thumbnail_url}


class RelatedVideo(Video):
    """"""
